#!/usr/bin/env python3
# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for models and editor representing the `slurm.conf` configuration file."""


from constants import EXAMPLE_SLURM_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import (
    DownNodes,
    FrontendNode,
    Node,
    NodeSet,
    Partition,
    SlurmConfig,
    slurmconfig,
)

EXPECTED_SLURM_DUMPS_OUTPUT = """
slurmctldhost=juju-c9fc6f-0(10.152.28.20)
slurmctldhost=juju-c9fc6f-1(10.152.28.100)
clustername=charmed-hpc
authtype=auth/munge
epilog=/usr/local/slurm/epilog
prolog=/usr/local/slurm/prolog
firstjobid=65536
inactivelimit=120
jobcomptype=jobcomp/filetxt
jobcomploc=/var/log/slurm/jobcomp
killwait=30
maxjobcount=10000
minjobage=3600
plugindir=/usr/local/lib:/usr/local/slurm/lib
returntoservice=0
schedulertype=sched/backfill
slurmctldlogfile=/var/log/slurm/slurmctld.log
slurmdlogfile=/var/log/slurm/slurmd.log
slurmctldport=7002
slurmdport=7003
slurmdspooldir=/var/spool/slurmd.spool
statesavelocation=/var/spool/slurm.state
tmpfs=/tmp
waittime=30
nodename=juju-c9fc6f-2 nodeaddr=10.152.28.48 cpus=1 realmemory=1000 tmpdisk=10000
nodename=juju-c9fc6f-3 nodeaddr=10.152.28.49 cpus=1 realmemory=1000 tmpdisk=10000
nodename=juju-c9fc6f-4 nodeaddr=10.152.28.50 cpus=1 realmemory=1000 tmpdisk=10000
nodename=juju-c9fc6f-5 nodeaddr=10.152.28.51 cpus=1 realmemory=1000 tmpdisk=10000
downnodes=juju-c9fc6f-5 state=down reason="Maintenance Mode"
partitionname=DEFAULT maxtime=30 maxnodes=10 state=up
partitionname=batch nodes=juju-c9fc6f-2,juju-c9fc6f-3,juju-c9fc6f-4,juju-c9fc6f-5 minnodes=4 maxtime=120 allowgroups=admin
""".strip()


class TestSlurmConfig(TestCase):
    """Unit tests for models and the editor representing the `slurm.conf` configuration file."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/slurm.conf", contents=EXAMPLE_SLURM_CONFIG)

    def test_loads(self) -> None:
        """Test `loads` method of `slurm.conf` model editor."""
        config = slurmconfig.loads(EXAMPLE_SLURM_CONFIG)
        self.assertListEqual(
            config.slurmctld_host,
            ["juju-c9fc6f-0(10.152.28.20)", "juju-c9fc6f-1(10.152.28.100)"],
        )

        self.assertEqual(config.slurmd_spool_dir, "/var/spool/slurmd.spool")
        self.assertEqual(config.scheduler_type, "sched/backfill")

        for name, node in config.nodes.items():
            self.assertIn(  # codespell:ignore
                name,
                {"juju-c9fc6f-2", "juju-c9fc6f-3", "juju-c9fc6f-4", "juju-c9fc6f-5"},
            )
            self.assertIn(  # codespell:ignore
                node.node_addr,
                {"10.152.28.48", "10.152.28.49", "10.152.28.50", "10.152.28.51"},
            )
            self.assertEqual(node.cpus, 1)
            self.assertEqual(node.real_memory, 1000)
            self.assertEqual(node.tmp_disk, 10000)

        for entry in config.down_nodes:
            self.assertEqual(entry.down_nodes[0], "juju-c9fc6f-5")
            self.assertEqual(entry.state, "down")
            self.assertEqual(entry.reason, "Maintenance Mode")

        for partition in config.partitions:
            self.assertIn(partition, {"DEFAULT", "batch"})  # codespell:ignore

        batch = config.partitions["batch"]
        self.assertListEqual(
            batch.nodes, ["juju-c9fc6f-2", "juju-c9fc6f-3", "juju-c9fc6f-4", "juju-c9fc6f-5"]
        )

    def test_dumps(self) -> None:
        """Test `dumps` method of the `slurm.conf` model editor."""
        config = slurmconfig.loads(EXAMPLE_SLURM_CONFIG)
        # Check if output matches expected output.
        self.assertEqual(slurmconfig.dumps(config), EXPECTED_SLURM_DUMPS_OUTPUT)
        # New config and old config should not be equal since the editor strips all comments.
        self.assertNotEqual(slurmconfig.dumps(config), EXAMPLE_SLURM_CONFIG)

    def test_edit(self) -> None:
        """Test `edit` context manager of the `slurm.conf` model editor."""
        with slurmconfig.edit("/etc/slurm/slurm.conf") as config:
            del config.inactive_limit
            config.reboot_program = "/usr/bin/systemctl reboot"
            config.max_job_count = 20000
            config.proctrack_type = "proctrack/linuxproc"
            config.plugin_dir.append("/snap/slurm/current/plugins")
            config.slurmctld_parameters = {"enable_configless": True}
            config.include = ["slurm.conf.overrides", "slurm.conf.profiling"]

            new_node = Node(**config.nodes["juju-c9fc6f-2"].dict())
            new_node.node_name = "batch-0"
            del config.nodes["juju-c9fc6f-2"]
            config.nodes[new_node.node_name] = new_node

        config = slurmconfig.load("/etc/slurm/slurm.conf")
        self.assertIsNone(config.inactive_limit)
        self.assertEqual(config.reboot_program, "/usr/bin/systemctl reboot")
        self.assertEqual(config.max_job_count, 20000)
        self.assertEqual(config.proctrack_type, "proctrack/linuxproc")
        self.assertListEqual(
            config.plugin_dir,
            ["/usr/local/lib", "/usr/local/slurm/lib", "/snap/slurm/current/plugins"],
        )
        self.assertDictEqual(config.slurmctld_parameters, {"enable_configless": True})
        self.assertListEqual(config.include, ["slurm.conf.overrides", "slurm.conf.profiling"])
        self.assertEqual(config.nodes["batch-0"].node_addr, "10.152.28.48")

        with slurmconfig.edit("/etc/slurm/slurm.conf") as config:
            del config.nodes
            del config.frontend_nodes
            del config.down_nodes
            del config.nodesets
            del config.partitions

        config = slurmconfig.load("/etc/slurm/slurm.conf")
        self.assertDictEqual(config.nodes.dict(), {})
        self.assertDictEqual(config.frontend_nodes.dict(), {})
        self.assertListEqual(config.down_nodes.list(), [])
        self.assertDictEqual(config.nodesets.dict(), {})
        self.assertDictEqual(config.partitions.dict(), {})

        new_nodes = [
            Node.from_dict(
                {
                    "nodename": "juju-c9fc6f-2",
                    "nodeaddr": "10.152.28.48",
                    "cpus": 1,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                }
            ),
            Node.from_dict(
                {
                    "nodename": "juju-c9fc6f-3",
                    "nodeaddr": "10.152.28.49",
                    "cpus": 1,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                }
            ),
            Node.from_dict(
                {
                    "nodename": "juju-c9fc6f-4",
                    "nodeaddr": "10.152.28.50",
                    "cpus": 1,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                }
            ),
            Node.from_dict(
                {
                    "nodename": "juju-c9fc6f-5",
                    "nodeaddr": "10.152.28.51",
                    "cpus": 1,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                }
            ),
        ]
        new_down_nodes = [
            DownNodes.from_dict(
                {
                    "downnodes": ["juju-c9fc6f-5"],
                    "state": "down",
                    "reason": "Maintenance Mode",
                }
            )
        ]
        new_partitions = [
            Partition.from_dict(
                {"partitionname": "DEFAULT", "maxtime": 30, "maxnodes": 10, "state": "up"}
            ),
            Partition.from_dict(
                {
                    "partitionname": "batch",
                    "nodes": [
                        "juju-c9fc6f-2",
                        "juju-c9fc6f-3",
                        "juju-c9fc6f-4",
                        "juju-c9fc6f-5",
                    ],
                    "minnodes": 4,
                    "maxtime": 120,
                    "allowgroups": ["admin"],
                }
            ),
        ]

        with slurmconfig.edit("/etc/slurm/slurm.conf") as config:
            for node in new_nodes:
                config.nodes[node.node_name] = node

            for down_node in new_down_nodes:
                config.down_nodes.append(down_node)

            for partition in new_partitions:
                config.partitions[partition.partition_name] = partition

    def test_update(self):
        """Test `update` method of the slurmconfig module."""
        config_updates = {
            "killwait": 10,
            "plugindir": ["/var/snap/slurm/usr/local/lib", "/var/snap/slurm/usr/local/slurm/lib"],
            "returntoservice": 0,
            "schedulertype": "sched/builtin",
            "switchtype": "switch/hpe_slingshot",
            "waittime": 30,
            "nodes": {
                "juju-c9fc6f-2": {
                    "nodeaddr": "10.152.28.98",
                    "cpus": 9,
                    "realmemory": 9000,
                    "tmpdisk": 90000,
                },
                "juju-c9fc6f-6": {
                    "nodeaddr": "10.152.28.52",
                    "cpus": 9,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                },
                "juju-c9fc6f-7": {
                    "nodeaddr": "10.152.28.53",
                    "cpus": 9,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                },
            },
            "downnodes": [
                {
                    "downnodes": ["juju-c9fc6f-6", "juju-c9fc6f-7"],
                    "state": "down",
                    "reason": "New nodes",
                }
            ],
            "partitions": {
                "DEFAULT": {
                    "maxtime": 10,
                    "maxnodes": 5,
                    "state": "up",
                },
                "new_batch": {
                    "nodes": ["juju-c9fc6f-6", "juju-c9fc6f-7"],
                    "minnodes": 1,
                    "maxtime": 120,
                    "allowgroups": ["admin"],
                },
            },
        }

        config = slurmconfig.loads(EXAMPLE_SLURM_CONFIG)
        updates = SlurmConfig.from_dict(config_updates)
        config.update(updates)

        self.assertEqual(config.kill_wait, 10)
        self.assertListEqual(
            config.plugin_dir,
            ["/var/snap/slurm/usr/local/lib", "/var/snap/slurm/usr/local/slurm/lib"],
        )
        self.assertEqual(config.scheduler_type, "sched/builtin")
        self.assertEqual(config.switch_type, "switch/hpe_slingshot")
        self.assertEqual(config.wait_time, 30)

        self.assertDictEqual(
            config.nodes.dict(),
            {
                "juju-c9fc6f-2": {
                    "nodeaddr": "10.152.28.98",
                    "cpus": 9,
                    "realmemory": 9000,
                    "tmpdisk": 90000,
                },
                "juju-c9fc6f-6": {
                    "nodeaddr": "10.152.28.52",
                    "cpus": 9,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                },
                "juju-c9fc6f-7": {
                    "nodeaddr": "10.152.28.53",
                    "cpus": 9,
                    "realmemory": 1000,
                    "tmpdisk": 10000,
                },
            },
        )
        self.assertListEqual(
            config.down_nodes.list(),
            [
                {
                    "downnodes": ["juju-c9fc6f-6", "juju-c9fc6f-7"],
                    "state": "down",
                    "reason": "New nodes",
                },
            ],
        )
        self.assertDictEqual(
            config.partitions.dict(),
            {
                "DEFAULT": {
                    "maxtime": 10,
                    "maxnodes": 5,
                    "state": "up",
                },
                "new_batch": {
                    "nodes": ["juju-c9fc6f-6", "juju-c9fc6f-7"],
                    "minnodes": 1,
                    "maxtime": 120,
                    "allowgroups": ["admin"],
                },
            },
        )

    def test_edit_fail(self) -> None:
        """Test that bogus attributes are handled when editing a loaded model."""
        # Catch if user tries to edit a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = slurmconfig.loads(EXAMPLE_SLURM_CONFIG)
            config.auth = "slurm"

        # Catch if user tries to access a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = slurmconfig.loads(EXAMPLE_SLURM_CONFIG)
            _ = config.auth

    def test_empty_config(self) -> None:
        """Test that the `slurm.conf` configuration file models correctly handle empty config."""
        config = SlurmConfig.from_str("")
        self.assertIsNone(config.accounting_storage_port)

        config = DownNodes.from_str("")
        self.assertIsNone(config.down_nodes)

        config = FrontendNode.from_str("")
        self.assertIsNone(config.deny_users)

        config = Node.from_str("")
        self.assertIsNone(config.node_addr)

        config = NodeSet.from_str("")
        self.assertIsNone(config.feature)

        config = Partition.from_str("")
        self.assertIsNone(config.partition_name)



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Callbacks for parsing and marshalling Slurm configuration models."""

__all__ = [
    "Callback",
    "DefaultCallback",
    "MultilineCallback",
    "BoolCallback",
    "StrBoolCallback",
    "IntBoolCallback",
    "QuoteCallback",
    "CommaDictCallback",
    "CommaDictColonPairCallback",
    "CommaDictColonArrayCallback",
    "SemicolonDictCallback",
    "CommaSepCallback",
    "ColonSepCallback",
    "make_callback",
]

import ast
from collections.abc import Callable, Iterable, Mapping
from functools import partial
from typing import Any, NamedTuple


class Callback(NamedTuple):
    """Callbacks for parsing and marshalling Slurm data model values.

    Args:
        parser: Callback that parses the value as read in from Slurm configuration.
        marshaller: Callback that marshals the value back into a valid Slurm configuration value.
    """

    parser: Callable[[str, str], Any]
    marshaller: Callable[[str, Any, str], str]


def make_callback(parser: Callable[[str], Any], marshaller: Callable[[Any], str]) -> Callback:
    """Make a new callback.

    Args:
        parser: Parser to initialize callback with.
        marshaller: Marshaller to initialize callback with.
    """
    return Callback(partial(_parse_base, parser), partial(_marshal_base, marshaller))


def _parse_base(parser: Callable[[str], Any], /, expr: str, sep: str) -> Any:
    """Base parser for defining Slurm configuration data parsers.

    Args:
        parser: Parser to handle value parsing and type casting.
        expr: Slurm configuration expression to parse.
        sep: Key, value separator used in Slurm configuration expression.
    """  # noqa D401
    _, v = expr.split(sep, maxsplit=1)
    return parser(v)


def _marshal_base(marshaller: Callable[[Any], str], /, key: str, value: Any, sep: str) -> str:
    """Base marshaller for defining Slurm configuration data marshallers.

    Args:
        marshaller: Marshaller to handle value marshalling and serialization.
        key: Configuration key to marshal.
        value: Configuration value to marshal.
        sep: Separator used to separate key and value in Slurm configuration.
    """  # noqa D401
    v = marshaller(value)
    # Split on newlines for non-unique configuration fields that can
    # appear more than once in a configuration file.
    return "\n".join(f"{key}{sep}{v_}" for v_ in v.strip().splitlines())


def _autocast(value: str) -> Any:
    """Autocast parsed configuration value to the 'best-fitting' Python type.

    Args:
        value: Parsed configuration value to automatically cast to the best-fitting Python type.
    """
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError):
        return value


def _bool_parse(value: str) -> bool:
    match value.lower():
        case "true":
            return True
        case "false":
            return False
        case _:
            raise ValueError(f"expected true or false, not {value}")


def _bool_marshal(value: bool) -> str:
    return "true" if value else "false"


def _strbool_parse(value: str) -> bool:
    match value.lower():
        case "yes":
            return True
        case "no":
            return False
        case _:
            raise ValueError(f"expected yes or no, not {value}")


def _strbool_marshal(value: bool) -> str:
    return "yes" if value else "no"


def _intbool_parse(value: str) -> bool:
    match value:
        case "1":
            return True
        case "0":
            return False
        case _:
            raise ValueError(f"expected 0 or 1, not {value}")


def _intbool_marshal(value: bool) -> str:
    return "1" if value else "0"


def _make_sep_callback(sep: str) -> Callback:
    r"""Make callback based on a given separator character.

    Args:
        sep: Separator character to generate parser and marshaller with.

    Examples:
        `_make_sep_callback(":")`
            Make a callback that will split a string on a colon (":")

        `_make_sep_callback(",")`
            Make a callback that will split a string on a comma (",")
    """

    def _sep_parser(value: str) -> list[Any]:
        return [_autocast(v) for v in value.split(sep)]

    def _sep_marshaller(value: Iterable[Any]) -> str:
        return sep.join(str(v) for v in value)

    return make_callback(_sep_parser, _sep_marshaller)


def _make_dict_callback(sep: str, pair_sep: str = "=", array_sep: str | None = None) -> Callback:
    r"""Make a callback based on given dictionary separators.

    Args:
        sep: Separator character used for separating key, value pairs.
        pair_sep: Separator character used for separating keys and values.
        array_sep: Separator character used for lists/arrays.

    Examples:
        `_make_dict_callback(",", "=", ",")`:
            Make a callback that will split the dictionary on a comma (","),
            split the key value pairs on equals sign ("="), and split array
            values on a comma (",").
    """

    def _dict_parser(value: str) -> dict[str, Any]:
        result: dict[str, Any] = {}

        for pair in value.split(sep):
            if pair_sep not in pair:
                result[pair] = True
                continue

            k, v = pair.split(pair_sep, maxsplit=1)
            if array_sep is not None and array_sep in v:
                result[k] = [_autocast(v_) for v_ in v.split(array_sep)]
            else:
                result[k] = _autocast(v)

        return result

    def _dict_marshaller(value: Mapping[str, Any]) -> str:
        result: list[str] = []

        for k, v in value.items():
            if isinstance(v, bool):
                # If `v` is `False`, omit `k` from final marshaled output.
                if v:
                    result.append(k)

                continue

            if array_sep is not None and isinstance(v, Iterable) and not isinstance(v, str):
                v = array_sep.join(str(v_) for v_ in v)

            result.append(f"{k}{pair_sep}{v}")

        return sep.join(result)

    return make_callback(_dict_parser, _dict_marshaller)


DefaultCallback = make_callback(_autocast, lambda v: str(v))
QuoteCallback = make_callback(_autocast, lambda v: f'"{v}"')
MultilineCallback = make_callback(_autocast, lambda v: "\n".join(v))
BoolCallback = make_callback(_bool_parse, _bool_marshal)
StrBoolCallback = make_callback(_strbool_parse, _strbool_marshal)
IntBoolCallback = make_callback(_intbool_parse, _intbool_marshal)
CommaDictCallback = _make_dict_callback(",", "=", ",")
CommaDictColonPairCallback = _make_dict_callback(",", ":", ",")
CommaDictColonArrayCallback = _make_dict_callback(",", "=", ":")
SemicolonDictCallback = _make_dict_callback(";", "=", ",")
ColonSepCallback = _make_sep_callback(":")
CommaSepCallback = _make_sep_callback(",")



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Model representing the `cgroup.conf` configuration file."""

__all__ = ["CGroupConfig", "CGroupConfigEditor"]

from typing import Annotated, Any

from .core.base import Metadata, Model, classproperty
from .core.callback import StrBoolCallback
from .core.editor import BaseEditor
from .core.schema import CGROUP_CONFIG_MODEL_SCHEMA


class CGroupConfig(Model):
    """Model representing the `cgroup.conf` configuration file.."""

    cgroup_mountpoint: str | None
    cgroup_plugin: str | None
    systemd_timeout: int | None
    ignore_systemd: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    ignore_systemd_on_failure: Annotated[bool, Metadata(callback=StrBoolCallback)]
    enable_controllers: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    allowed_ram_space: float | None
    allowed_swap_space: float | None
    constrain_cores: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    constrain_devices: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    constrain_ram_space: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    constrain_swap_space: Annotated[bool | None, Metadata(callback=StrBoolCallback)]
    max_ram_percent: float | None
    max_swap_percent: float | None
    memory_swappiness: int | None
    min_ram_space: float | None
    signal_children_processes: Annotated[bool | None, Metadata(callback=StrBoolCallback)]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N805
        return CGROUP_CONFIG_MODEL_SCHEMA


class CGroupConfigEditor(BaseEditor):
    """Editor for the `cgroup.conf` configuration file."""

    @property
    def __model__(self) -> type[Model]:  # noqa D105
        return CGroupConfig



#!/usr/bin/env python3
# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the model and editor of the `oci.conf` configuration file."""

from constants import EXAMPLE_OCI_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import ModelError, OCIConfig, ociconfig

EXPECTED_OCI_DUMPS_OUTPUT = """
ignorefileconfigjson=true
envexclude="^(SLURM_CONF|SLURM_CONF_SERVER)="
runtimeenvexclude="^(SLURM_CONF|SLURM_CONF_SERVER)="
runtimerun="singularity exec --userns %r %@"
runtimekill="kill -s SIGTERM %p"
runtimedelete="kill -s SIGKILL %p"
""".strip()


class TestOCIConfig(TestCase):
    """Unit tests for the model and editor of the `oci.conf` configuration file."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/oci.conf", contents=EXAMPLE_OCI_CONFIG)

    def test_loads(self) -> None:
        """Test `loads` method of the `oci.conf` model editor."""
        config = ociconfig.loads(EXAMPLE_OCI_CONFIG)
        self.assertTrue(config.ignore_file_config_json)
        self.assertEqual(config.env_exclude, "^(SLURM_CONF|SLURM_CONF_SERVER)=")
        self.assertEqual(config.run_time_env_exclude, "^(SLURM_CONF|SLURM_CONF_SERVER)=")
        self.assertEqual(config.run_time_run, "singularity exec --userns %r %@")
        self.assertEqual(config.run_time_kill, "kill -s SIGTERM %p")
        self.assertEqual(config.run_time_delete, "kill -s SIGKILL %p")

    def test_dumps(self) -> None:
        """Test `dumps` method of the `oci.conf` model editor."""
        config = ociconfig.loads(EXAMPLE_OCI_CONFIG)
        # Check if output matches expected output.
        self.assertEqual(ociconfig.dumps(config), EXPECTED_OCI_DUMPS_OUTPUT)
        # New config and old config should not be equal since the editor strips all comments.
        self.assertNotEqual(ociconfig.dumps(config), EXAMPLE_OCI_CONFIG)

    def test_edit(self) -> None:
        """Test `edit` context manager of the `oci.conf` model editor."""
        with ociconfig.edit("/etc/slurm/oci.conf") as config:
            config.ignore_file_config_json = False
            config.env_exclude = "^(SLURM_CONF)="
            config.run_time_run = "sudo " + config.run_time_run
            config.run_time_kill = "sudo " + config.run_time_kill
            config.run_time_delete = "sudo " + config.run_time_delete

        config = ociconfig.load("/etc/slurm/oci.conf")
        self.assertFalse(config.ignore_file_config_json)
        self.assertEqual(config.env_exclude, "^(SLURM_CONF)=")
        self.assertEqual(config.run_time_run, "sudo singularity exec --userns %r %@")
        self.assertEqual(config.run_time_kill, "sudo kill -s SIGTERM %p")
        self.assertEqual(config.run_time_delete, "sudo kill -s SIGKILL %p")

    def test_load_fail(self) -> None:
        """Test that bogus values are automatically caught when loading a new model."""
        # Catch if field value is unexpected. e.g. not a boolean value.
        with self.assertRaises(ValueError):
            OCIConfig.from_str(
                """
                ignorefileconfigjson=totallybro
                """
            )

        # Catch if field is unexpected or in the incorrect syntax.
        with self.assertRaises(ModelError):
            OCIConfig.from_str(
                """
                run_time_kill="kill -9 %p"
                """
            )

        # Catch if field value is invalid when checking against the model schema.
        with self.assertRaises(ModelError):
            OCIConfig.from_str(
                """
                createenvfile=nodontdothat
                """
            )

    def test_edit_fail(self) -> None:
        """Test that bogus attributes are handled when editing a loaded model."""
        # Catch if user tries to edit a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = ociconfig.loads(EXAMPLE_OCI_CONFIG)
            config.env_include = "^(SLURM_JWT)="

        # Catch if user tries to access a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = ociconfig.loads(EXAMPLE_OCI_CONFIG)
            _ = config.env_include



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the `utils` module."""

from unittest import TestCase

from slurmutils import calculate_rs


class TestUtils(TestCase):
    """Unit tests for the `utils` module ."""

    def test_calculate_rs(self) -> None:
        """Test the `calculate_rs` utility function."""
        node_range = [0, 1, 2, 3, 4, 5, 6, 8, 9, 10, 12, 14, 15, 16, 18]
        r = calculate_rs(node_range)
        self.assertEqual(r, "[0-6,8-10,12,14-16,18]")



#!/usr/bin/env python3
# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for base editor functions."""

import os
import stat
from pathlib import Path

from constants import EXAMPLE_SLURM_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import slurmconfig
from slurmutils.core.editor import _set_file_permissions  # noqa


class TestBaseEditor(TestCase):
    """Unit tests for base editor functions."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/slurm.conf", contents=EXAMPLE_SLURM_CONFIG)

    def test_set_file_permissions(self) -> None:
        """Test the `set_file_permissions` function."""
        target = Path("/etc/slurm/slurm.conf")
        _set_file_permissions(target, mode=0o600, user=os.getuid(), group=os.getgid())
        f_info = target.stat()
        self.assertEqual("-rw-------", stat.filemode(f_info.st_mode))
        self.assertEqual(os.getuid(), f_info.st_uid)
        self.assertEqual(os.getgid(), f_info.st_gid)

    def test_loader_fail(self) -> None:
        """Test that `FileNotFoundError` is raised when attempting to load non-existent file."""
        self.fs.remove("/etc/slurm/slurm.conf")
        with self.assertRaises(FileNotFoundError):
            slurmconfig.load("/etc/slurm/slurm.conf")

    def test_dumper_first_write(self) -> None:
        """Test that `dumper` succeeds when there is no pre-existing config file."""
        self.fs.remove("/etc/slurm/slurm.conf")
        slurmconfig.dump(slurmconfig.loads(EXAMPLE_SLURM_CONFIG), "/etc/slurm/slurm.conf")



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

from .core.base import Model, classproperty
from .core.editor import BaseEditor

class OCIConfig(Model):
    container_path: str | None
    create_env_file: str | None
    debug_flags: list[str] | None
    disable_cleanup: bool | None
    disable_hooks: list[str] | None
    env_exclude: str | None
    mount_spool_dir: str | None
    run_time_env_exclude: str | None
    file_debug: str | None
    ignore_file_config_json: bool | None
    run_time_create: str | None
    run_time_delete: str | None
    run_time_kill: str | None
    run_time_query: str | None
    run_time_run: str | None
    run_time_start: str | None
    srun_path: str | None
    srun_args: list[str] | None
    std_io_debug: str | None
    syslog_debug: str | None
    @classproperty
    def __model_schema__(self) -> dict[str, Any]: ...  # noqa N804

class OCIConfigEditor(BaseEditor):
    @property
    def __model__(self) -> type[OCIConfig]: ...
    def dump(
        self,
        obj: OCIConfig,
        /,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: str | int | None = None,
        group: str | int | None = None,
    ) -> None: ...
    def dumps(self, obj: OCIConfig, /) -> str: ...
    def load(self, file: str | os.PathLike, /) -> OCIConfig: ...
    def loads(self, s: str, /) -> OCIConfig: ...
    @contextmanager
    def edit(
        self,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: int | str | None = None,
        group: int | str | None = None,
    ) -> Iterator[OCIConfig]: ...



#!/usr/bin/env python3
# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the model and editor of the `cgroup.conf` configuration file."""


from constants import EXAMPLE_CGROUP_CONFIG
from pyfakefs.fake_filesystem_unittest import TestCase

from slurmutils import CGroupConfig, ModelError, cgroupconfig

EXPECTED_CGROUP_DUMPS_OUTPUT = """
constraincores=yes
constraindevices=yes
constrainramspace=yes
constrainswapspace=yes
""".strip()


class TestCGroupConfig(TestCase):
    """Unit tests for model and editor of the `cgroup.conf` configuration file."""

    def setUp(self) -> None:
        self.setUpPyfakefs()
        self.fs.create_file("/etc/slurm/cgroup.conf", contents=EXAMPLE_CGROUP_CONFIG)

    def test_loads(self) -> None:
        """Test `loads` method of the `cgroup.conf` model editor."""
        config = cgroupconfig.loads(EXAMPLE_CGROUP_CONFIG)
        self.assertTrue(config.constrain_cores)
        self.assertTrue(config.constrain_devices)
        self.assertTrue(config.constrain_ram_space)
        self.assertTrue(config.constrain_swap_space)

    def test_dumps(self) -> None:
        """Test `dumps` method of the `cgroup.conf` model editor."""
        config = cgroupconfig.loads(EXAMPLE_CGROUP_CONFIG)
        # Check if output matches expected output.
        self.assertEqual(cgroupconfig.dumps(config), EXPECTED_CGROUP_DUMPS_OUTPUT)
        # New config and old config should not be equal since the editor strips all comments.
        self.assertNotEqual(cgroupconfig.dumps(config), EXAMPLE_CGROUP_CONFIG)

    def test_edit(self) -> None:
        """Test `edit` context manager of the `cgroup.conf` model editor."""
        with cgroupconfig.edit("/etc/slurm/cgroup.conf") as config:
            config.constrain_cores = False
            config.constrain_devices = False
            config.constrain_ram_space = False
            config.constrain_swap_space = False

        config = cgroupconfig.load("/etc/slurm/cgroup.conf")
        self.assertFalse(config.constrain_cores)
        self.assertFalse(config.constrain_devices)
        self.assertFalse(config.constrain_ram_space)
        self.assertFalse(config.constrain_swap_space)

    def test_load_fail(self) -> None:
        """Test that bogus values are automatically caught when loading a new model."""
        # Catch if field value is unexpected. e.g. not a boolean value.
        with self.assertRaises(ValueError):
            CGroupConfig.from_str(
                """
                constrainramspace=totally
                """
            )

        # Catch if field is unexpected or in the incorrect syntax.
        with self.assertRaises(ModelError):
            CGroupConfig.from_str(
                """
                constrain_ram_space=yes
                """
            )

        # Catch if field value is invalid when checking against the model schema.
        with self.assertRaises(ModelError):
            CGroupConfig.from_str(
                """
                cgroupplugin=yeahtotallyusecgroup
                """
            )

    def test_edit_fail(self) -> None:
        """Test that bogus attributes are handled when editing a loaded model."""
        # Catch if user tries to edit a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = cgroupconfig.loads(CGroupConfig)
            config.constrain_ram = False

        # Catch if user tries to access a non-existent attribute.
        with self.assertRaises(AttributeError):
            config = cgroupconfig.loads(CGroupConfig)
            _ = config.constrain_ram



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Exceptions raised by Slurm utilities in this package."""


class BaseError(Exception):
    """Base exception for errors in `slurmutils` module."""

    @property
    def message(self) -> str:
        """Return message passed as argument to exception."""
        return self.args[0]


class ModelError(BaseError):
    """Raise when a Slurm configuration model encounters an error."""



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Any

from .core.base import Mode, Model, ModelList, ModelMapping, classproperty
from .core.callback import Callback
from .core.editor import BaseEditor

MCSParametersCallback: Callback = ...

class DownNodes(Model):
    down_nodes: list[str] | None
    reason: str | None
    state: str | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_mode__(cls) -> Mode: ...  # noqa N804

class FrontendNode(Model):
    frontend_name: str | None
    frontend_addr: str | None
    allow_groups: list[str] | None
    allow_users: list[str] | None
    deny_groups: list[str] | None
    deny_users: list[str] | None
    port: int | None
    reason: str | None
    state: str | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_mode__(cls) -> Mode: ...  # noqa N804

class Node(Model):
    node_name: str | None
    node_hostname: str | None
    node_addr: str | None
    bcast_addr: str | None
    boards: int | None
    core_spec_count: int | None
    cores_per_socket: int | None
    cpu_bind: str | None
    cpus: int | None
    cpu_spec_list: list[int] | None
    features: list[str] | None
    gres: list[str] | None
    mem_spec_limit: int | None
    port: int | None
    procs: int | None
    real_memory: int | None
    reason: str | None
    sockets: int | None
    sockets_per_board: int | None
    state: str | None
    threads_per_core: int | None
    tmp_disk: int | None
    weight: int | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_mode__(cls) -> Mode: ...  # noqa N804

class NodeSet(Model):
    node_set: str | None
    feature: str | None
    nodes: list[str] | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_mode__(cls) -> Mode: ...  # noqa N804

class Partition(Model):
    partition_name: str | None
    alloc_nodes: list[str] | None
    allow_accounts: list[str] | None
    allow_groups: list[str] | None
    allow_qos: list[str] | None
    alternate: str | None
    cpu_bind: str | None
    default: bool | None
    default_time: int | str | None
    def_cpu_per_gpu: int | None
    def_mem_per_cpu: int | None
    def_mem_per_gpu: int | None
    def_mem_per_node: int | None
    deny_accounts: list[str] | None
    deny_qos: list[str] | None
    disable_root_jobs: bool | None
    exclusive_topo: bool | None
    exclusive_user: bool | None
    grace_time: int | None
    hidden: bool | None
    lln: bool | None
    max_cpus_per_node: int | None
    max_cpus_per_socket: int | None
    max_mem_per_cpu: int | None
    max_mem_per_node: int | None
    max_nodes: int | str | None
    max_time: int | str | None
    min_nodes: int | None
    nodes: list[str] | None
    over_subscribe: str | None
    over_time_limit: int | str | None
    power_down_on_idle: bool | None
    preempt_mode: str | None
    priority_job_factor: int | None
    priority_tier: int | None
    qos: str | None
    req_resv: bool | None
    resume_timeout: int | None
    root_only: bool | None
    select_type_parameters: list[str] | None
    state: str | None
    suspend_time: int | str | None
    suspend_timeout: int | None
    tres_billing_weights: dict[str, str] | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_mode__(cls) -> Mode: ...  # noqa N804

class DownNodesList(ModelList[DownNodes]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]: ...  # noqa N804

class FrontendNodeMapping(ModelMapping[str, FrontendNode]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]: ...  # noqa N804

class NodeMapping(ModelMapping[str, Node]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]: ...  # noqa N804

class NodeSetMapping(ModelMapping[str, NodeSet]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]: ...  # noqa N804

class PartitionMapping(ModelMapping[str, Partition]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]: ...  # noqa N804

class SlurmConfig(Model):
    accounting_storage_backup_host: str | None
    accounting_storage_enforce: list[str] | None
    accounting_storage_external_host: str | None
    accounting_storage_host: str | None
    accounting_storage_parameters: dict[str, str | list[str]] | None
    accounting_storage_pass: str | None
    accounting_storage_port: int | None
    accounting_storage_tres: list[str] | None
    accounting_storage_type: str | None
    accounting_storage_user: str | None
    accounting_store_flags: list[str] | None
    acct_gather_node_freq: int | None
    acct_gather_energy_type: str | None
    acct_gather_interconnect_type: str | None
    acct_gather_filesystem_type: str | None
    acct_gather_profile_type: str | None
    allow_spec_resources_usage: bool | None
    auth_alt_types: list[str] | None
    auth_alt_parameters: dict[str, bool | int | str] | None
    auth_info: dict[str, bool | int | str] | None
    auth_type: str | None
    batch_start_timeout: int | None
    bcast_exclude: list[str] | None
    bcast_parameters: dict[str, bool | str] | None
    burst_buffer_type: str | None
    cli_filter_plugins: list[str] | None
    cluster_name: str | None
    communication_parameters: dict[str, bool | int] | None
    complete_wait: int | None
    cpu_freq_def: str | None
    cpu_freq_governors: list[str] | None
    cred_type: str | None
    debug_flags: list[str] | None
    def_cpu_per_gpu: int | None
    def_mem_per_cpu: int | None
    def_mem_per_gpu: int | None
    def_mem_per_node: int | None
    dependency_parameters: dict[str, bool | int] | None
    disable_root_jobs: bool | None
    eio_timeout: int | None
    enforce_part_limits: str | None
    epilog: str | None
    epilog_msg_time: int | None
    epilog_slurmctld: str | None
    fair_share_dampening_factor: int | None
    federation_parameters: dict[str, bool] | None
    first_job_id: int | None
    get_env_timeout: int | None
    gres_types: list[str] | None
    group_update_force: bool | None
    group_update_time: int | None
    gpu_freq_def: int | str | None
    health_check_interval: int | None
    health_check_node_state: list[str] | None
    health_check_program: str | None
    inactive_limit: int | None
    include: list[str] | None
    interactive_step_options: str | None
    job_acct_gather_type: str | None
    job_acct_gather_frequency: dict[str, int] | None
    job_acct_gather_params: dict[str, bool] | None
    job_comp_host: str | None
    job_comp_loc: str | None
    job_comp_params: dict[str, bool | int | str] | None
    job_comp_pass: str | None
    job_comp_port: int | None
    job_comp_type: str | None
    job_comp_user: str | None
    job_container_type: str | None
    job_file_append: bool | None
    job_requeue: bool | None
    job_submit_plugins: list[str] | None
    kill_on_bad_exit: bool | None
    kill_wait: int | None
    max_batch_requeue: int | None
    node_features_plugins: list[str] | None
    launch_parameters: dict[str, bool] | None
    licenses: dict[str, bool | int] | None
    log_time_format: str | None
    mail_domain: str | None
    mail_prog: str | None
    max_array_size: int | None
    max_dbd_msgs: int | None
    max_job_count: int | None
    max_job_id: int | None
    max_mem_per_cpu: int | None
    max_mem_per_node: int | None
    max_node_count: int | None
    max_step_count: int | None
    max_tasks_per_node: int | None
    mcs_parameters: dict[str, bool | list[str]] | None
    mcs_plugin: str | None
    message_timeout: int | None
    min_job_age: int | None
    mpi_default: str | None
    mpi_params: dict[str, bool | str] | None
    over_time_limit: int | str | None
    plugin_dir: list[str] | None
    plug_stack_config: str | None
    preempt_mode: str | None
    preempt_parameters: dict[str, bool | int] | None
    preempt_type: str | None
    preempt_exempt_time: int | str | None
    pr_ep_parameters: dict[str, Any] | None
    pr_ep_plugins: list[str] | None
    priority_calc_period: int
    priority_decay_half_life: int | str | None
    priority_favor_small: bool | None
    priority_flags: list[str] | None
    priority_max_age: int | str | None
    priority_parameters: str | None
    priority_site_factor_parameters: str | None
    priority_site_factor_plugin: str | None
    priority_type: str | None
    priority_usage_reset_period: str | None
    priority_weight_age: int | None
    priority_weight_assoc: int | None
    priority_weight_fairshare: int | None
    priority_weight_job_size: int | None
    priority_weight_partition: int | None
    priority_weight_qos: int | None
    priority_weight_tres: dict[str, int] | None
    private_data: list[str] | None
    proctrack_type: str | None
    prolog: str | None
    prolog_epilog_timeout: int | None
    prolog_flags: list[str] | None
    prolog_slurmctld: str | None
    propagate_prio_process: int | None
    propagate_resource_limits: list[str] | None
    propagate_resource_limits_except: list[str] | None
    reboot_program: str | None
    reconfig_flags: list[str] | None
    requeue_exit: list[str | int] | None
    requeue_exit_hold: list[str | int] | None
    resume_fail_program: str | None
    resume_program: str | None
    resume_rate: int | None
    resume_timeout: int | None
    resv_epilog: str | None
    resv_over_run: int | str | None
    resv_prolog: str | None
    return_to_service: int | None
    scheduler_parameters: dict[str, bool | int | str] | None
    scheduler_time_slice: int | None
    scheduler_type: str | None
    scron_parameters: dict[str, bool] | None
    select_type: str | None
    select_type_parameters: dict[str, bool] | None
    slurmctld_addr: str | None
    slurmctld_debug: str | None
    slurmctld_host: list[str] | None
    slurmctld_log_file: str | None
    slurmctld_parameters: dict[str, bool | int | str] | None
    slurmctld_pid_file: str | None
    slurmctld_port: int | str | None
    slurmctld_primary_off_prog: str | None
    slurmctld_primary_on_prog: str | None
    slurmctld_syslog_debug: str | None
    slurmctld_timeout: int | None
    slurmd_debug: str | None
    slurmd_log_file: str | None
    slurmd_parameters: dict[str, bool | int] | None
    slurmd_pid_file: str | None
    slurmd_port: int | None
    slurmd_spool_dir: str | None
    slurmd_syslog_debug: str | None
    slurmd_timeout: int | None
    slurmd_user: str | None
    slurm_sched_log_file: str | None
    slurm_sched_log_level: bool | None
    slurm_user: str | None
    srun_epilog: str | None
    srun_port_range: str | None
    srun_prolog: str | None
    state_save_location: str | None
    suspend_exc_nodes: dict[str, bool | int] | None
    suspend_exc_parts: list[str] | None
    suspend_exc_states: list[str] | None
    suspend_program: str | None
    suspend_rate: int | None
    suspend_time: int | str | None
    suspend_timeout: int | None
    switch_parameters: dict[str, bool | int | str | list[str]] | None
    switch_type: str | None
    task_epilog: str | None
    task_plugin: list[str] | None
    task_plugin_param: dict[str, bool | str] | None
    task_prolog: str | None
    tcp_timeout: int | None
    tmpfs: str | None
    topology_param: dict[str, bool | int] | None
    topology_plugin: str | None
    track_wc_key: bool | None
    tree_width: int | None
    unkillable_step_program: str | None
    unkillable_step_timeout: int | None
    use_pam: bool | None
    v_size_factor: int | None
    wait_time: int | None
    x11_parameters: dict[str, bool] | None
    down_nodes: DownNodesList
    frontend_nodes: FrontendNodeMapping
    nodes: NodeMapping
    nodesets: NodeSetMapping
    partitions: PartitionMapping
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any]: ...  # noqa N804

class SlurmConfigEditor(BaseEditor):
    @property
    def __model__(self) -> type[SlurmConfig]: ...
    def dump(
        self,
        obj: SlurmConfig,
        /,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: str | int | None = None,
        group: str | int | None = None,
    ) -> None: ...
    def dumps(self, obj: SlurmConfig, /) -> str: ...
    def load(self, file: str | os.PathLike, /) -> SlurmConfig: ...
    def loads(self, s: str, /) -> SlurmConfig: ...
    @contextmanager
    def edit(
        self,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: int | str | None = None,
        group: int | str | None = None,
    ) -> Iterator[SlurmConfig]: ...

_down_nodes_list_model_builder: Callable[[Any], Any] = ...
_frontend_node_mapping_model_builder: Callable[[Any], Any] = ...
_node_mapping_model_builder: Callable[[Any], Any] = ...
_nodeset_mapping_model_builder: Callable[[Any], Any] = ...
_partition_mapping_model_builder: Callable[[Any], Any] = ...
_slurm_config_model_builder: Callable[[Any], Any] = ...



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Models representing the `gres.conf` configuration file."""

__all__ = ["Gres", "GresList", "GresMapping", "GresConfig", "GresConfigEditor"]

from typing import Annotated, Any, Callable

from .core.base import (
    Metadata,
    Mode,
    Model,
    ModelList,
    ModelMapping,
    classproperty,
    make_model_builder,
)
from .core.callback import CommaSepCallback
from .core.editor import BaseEditor
from .core.schema import (
    GRES_CONFIG_MODEL_SCHEMA,
    GRES_LIST_MODEL_SCHEMA,
    GRES_MAPPING_MODEL_SCHEMA,
    GRES_MODEL_SCHEMA,
)


class Gres(Model):
    """Model representing a generic resource declaration in the `gres.conf` configuration file."""

    name: Annotated[str | None, Metadata(primary=True)]
    autodetect: str | None
    count: int | str | None
    cores: Annotated[list[int | str] | None, Metadata(callback=CommaSepCallback)]
    file: str | None
    flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    links: Annotated[list[int] | None, Metadata(callback=CommaSepCallback)]
    multiplefiles: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    nodename: str | None
    type: str | None

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N804
        return GRES_MODEL_SCHEMA

    @classproperty
    def __model_mode__(cls) -> Mode:  # noqa N804
        return Mode.ONELINE


class GresList(ModelList[Gres]):
    """List of generic resource declarations that share the same name."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N804
        return GRES_LIST_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None:  # noqa N804
        return _gres_list_model_builder


class GresMapping(ModelMapping[str, GresList]):
    """Mapping of generic resource names to generic resource declarations."""

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N804
        return GRES_MAPPING_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None:  # noqa N804
        return _gres_mapping_model_builder


class GresConfig(Model):
    """Model representing the `gres.conf` configuration file."""

    auto_detect: str | None
    gres: Annotated[
        GresMapping,
        Metadata(origin="gres", alias="name", unique=False, default_factory=lambda: GresMapping()),
    ]

    @classproperty
    def __model_schema__(cls) -> dict[str, Any]:  # noqa N804
        return GRES_CONFIG_MODEL_SCHEMA

    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None:  # noqa N804
        return _gres_config_model_builder


class GresConfigEditor(BaseEditor):
    """Editor for the `gres.conf` configuration file."""

    @property
    def __model__(self) -> type[Model]:  # noqa D105
        return GresConfig


_gres_list_model_builder = make_model_builder(Gres)
_gres_mapping_model_builder = make_model_builder(GresList)
_gres_config_model_builder = make_model_builder(GresMapping)



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Any

from .core.base import Mode, Model, ModelList, ModelMapping, classproperty
from .core.editor import BaseEditor

class Gres(Model):
    name: str | None
    autodetect: str | None
    count: int | str | None
    cores: list[int | str] | None
    file: str | None
    flags: list[str] | None
    links: list[int] | None
    multiplefiles: list[str] | None
    nodename: str | None
    type: str | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_mode__(cls) -> Mode: ...  # noqa N804

class GresList(ModelList[Gres]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None: ...  # noqa N804

class GresMapping(ModelMapping[str, GresList]):
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None: ...  # noqa N804

class GresConfig(Model):
    auto_detect: str | None
    gres: GresMapping
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804
    @classproperty
    def __model_builder__(cls) -> Callable[[Any], Any] | None: ...  # noqa N804

class GresConfigEditor(BaseEditor):
    @property
    def __model__(self) -> type[GresConfig]: ...
    def dump(
        self,
        obj: GresConfig,
        /,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: str | int | None = None,
        group: str | int | None = None,
    ) -> None: ...
    def dumps(self, obj: GresConfig, /) -> str: ...
    def load(self, file: str | os.PathLike, /) -> GresConfig: ...
    def loads(self, s: str, /) -> GresConfig: ...
    @contextmanager
    def edit(
        self,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: int | str | None = None,
        group: int | str | None = None,
    ) -> Iterator[GresConfig]: ...

_gres_list_model_builder: Callable[[Any], Any] = ...
_gres_mapping_model_builder: Callable[[Any], Any] = ...
_gres_config_model_builder: Callable[[Any], Any] = ...



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Model representing the `oci.conf` configuration file."""

__all__ = ["OCIConfig", "OCIConfigEditor"]

from typing import Annotated, Any

from .core.base import Metadata, Model, classproperty
from .core.callback import (
    BoolCallback,
    CommaSepCallback,
    MultilineCallback,
    QuoteCallback,
)
from .core.editor import BaseEditor
from .core.schema import OCI_CONFIG_MODEL_SCHEMA


class OCIConfig(Model):
    """Model representing the `oci.conf` configuration file."""

    container_path: str | None
    create_env_file: str | None
    debug_flags: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    disable_cleanup: Annotated[bool | None, Metadata(callback=BoolCallback)]
    disable_hooks: Annotated[list[str] | None, Metadata(callback=CommaSepCallback)]
    env_exclude: Annotated[str | None, Metadata(callback=QuoteCallback)]
    mount_spool_dir: str | None
    run_time_env_exclude: Annotated[str | None, Metadata(callback=QuoteCallback)]
    file_debug: str | None
    ignore_file_config_json: Annotated[bool | None, Metadata(callback=BoolCallback)]
    run_time_create: Annotated[str | None, Metadata(callback=QuoteCallback)]
    run_time_delete: Annotated[str | None, Metadata(callback=QuoteCallback)]
    run_time_kill: Annotated[str | None, Metadata(callback=QuoteCallback)]
    run_time_query: Annotated[str | None, Metadata(callback=QuoteCallback)]
    run_time_run: Annotated[str | None, Metadata(callback=QuoteCallback)]
    run_time_start: Annotated[str | None, Metadata(callback=QuoteCallback)]
    srun_path: str | None
    srun_args: Annotated[list[str] | None, Metadata(unique=False, callback=MultilineCallback)]
    std_io_debug: str | None
    syslog_debug: str | None

    @classproperty
    def __model_schema__(self) -> dict[str, Any]:  # noqa N805
        return OCI_CONFIG_MODEL_SCHEMA


class OCIConfigEditor(BaseEditor):
    """Editor for the `oci.conf` configuration file."""

    @property
    def __model__(self) -> type[Model]:  # noqa D105
        return OCIConfig



# Copyright 2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

from .core.base import Model, classproperty
from .core.editor import BaseEditor

class AcctGatherConfig(Model):
    energy_ipmi_frequency: int | None
    energy_ipmi_calc_adjustment: bool | None
    energy_ipmi_power_sensors: dict[str, list[int]] | None
    energy_ipmi_username: str | None
    energy_ipmi_password: str | None
    energy_ipmi_frequency: int | None
    energy_ipmi_timeout: int | None
    profile_hdf5_dir: str | None
    profile_hdf5_default: list[str] | None
    profile_influxdb_database: str | None
    profile_influxdb_default: list[str] | None
    profile_influxdb_host: str | None
    profile_influxdb_pass: str | None
    profile_influxdb_rt_policy: str | None
    profile_influxdb_user: str | None
    profile_influxdb_timeout: int | None
    infiniband_ofed_port: int | None
    sysfs_interfaces: list[str] | None
    @classproperty
    def __model_schema__(cls) -> dict[str, Any]: ...  # noqa N804

class AcctGatherConfigEditor(BaseEditor):
    @property
    def __model__(self) -> type[AcctGatherConfig]: ...
    def dump(
        self,
        obj: AcctGatherConfig,
        /,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: str | int | None = None,
        group: str | int | None = None,
    ) -> None: ...
    def dumps(self, obj: AcctGatherConfig, /) -> str: ...
    def load(self, file: str | os.PathLike, /) -> AcctGatherConfig: ...
    def loads(self, s: str, /) -> AcctGatherConfig: ...
    @contextmanager
    def edit(
        self,
        file: str | os.PathLike,
        *,
        mode: int = 0o644,
        user: int | str | None = None,
        group: int | str | None = None,
    ) -> Iterator[AcctGatherConfig]: ...



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

EXAMPLE_GRES_CONFIG = """#
# `gres.conf` file generated at 2024-12-10 14:17:35.161642 by slurmutils.
#
AutoDetect=nvml
Name=gpu Type=gp100  File=/dev/nvidia0 Cores=0,1
Name=gpu Type=gp100  File=/dev/nvidia1 Cores=0,1
Name=gpu Type=p6000  File=/dev/nvidia2 Cores=2,3
Name=gpu Type=p6000  File=/dev/nvidia3 Cores=2,3
Name=mps Count=200  File=/dev/nvidia0
Name=mps Count=200  File=/dev/nvidia1
Name=mps Count=100  File=/dev/nvidia2
Name=mps Count=100  File=/dev/nvidia3
Name=bandwidth Type=lustre Count=4G Flags=countonly
Name=gpu NodeName=juju-abc654-1 Type=tesla_t4 File=/dev/nvidia[0-1] Count=8G
Name=gpu NodeName=juju-abc654-1 Type=l40s File=/dev/nvidia[2-3] Count=12G
"""

EXAMPLE_OCI_CONFIG = """#
# `oci.conf` file generated at 2024-09-18 19:05:45.723019 by slurmutils.
#
IgnoreFileConfigJson=true
EnvExclude="^(SLURM_CONF|SLURM_CONF_SERVER)="
RunTimeEnvExclude="^(SLURM_CONF|SLURM_CONF_SERVER)="
RunTimeRun="singularity exec --userns %r %@"
RunTimeKill="kill -s SIGTERM %p"
RunTimeDelete="kill -s SIGKILL %p"
"""

EXAMPLE_SLURM_CONFIG = """#
# `slurm.conf` file generated at 2024-01-30 17:18:36.171652 by slurmutils.
#
SlurmctldHost=juju-c9fc6f-0(10.152.28.20)
SlurmctldHost=juju-c9fc6f-1(10.152.28.100)

ClusterName=charmed-hpc
AuthType=auth/munge
Epilog=/usr/local/slurm/epilog
Prolog=/usr/local/slurm/prolog
FirstJobId=65536
InactiveLimit=120
JobCompType=jobcomp/filetxt
JobCompLoc=/var/log/slurm/jobcomp
KillWait=30
MaxJobCount=10000
MinJobAge=3600
PluginDir=/usr/local/lib:/usr/local/slurm/lib
ReturnToService=0
SchedulerType=sched/backfill
SlurmctldLogFile=/var/log/slurm/slurmctld.log
SlurmdLogFile=/var/log/slurm/slurmd.log
SlurmctldPort=7002
SlurmdPort=7003
SlurmdSpoolDir=/var/spool/slurmd.spool
StateSaveLocation=/var/spool/slurm.state
TmpFS=/tmp
WaitTime=30

#
# Node configurations
#
NodeName=juju-c9fc6f-2 NodeAddr=10.152.28.48 CPUs=1 RealMemory=1000 TmpDisk=10000
NodeName=juju-c9fc6f-3 NodeAddr=10.152.28.49 CPUs=1 RealMemory=1000 TmpDisk=10000
NodeName=juju-c9fc6f-4 NodeAddr=10.152.28.50 CPUs=1 RealMemory=1000 TmpDisk=10000
NodeName=juju-c9fc6f-5 NodeAddr=10.152.28.51 CPUs=1 RealMemory=1000 TmpDisk=10000

#
# Down node configurations
#
DownNodes=juju-c9fc6f-5 State=down Reason="Maintenance Mode"

#
# Partition configurations
#
PartitionName=DEFAULT MaxTime=30 MaxNodes=10 State=up
PartitionName=batch Nodes=juju-c9fc6f-2,juju-c9fc6f-3,juju-c9fc6f-4,juju-c9fc6f-5 MinNodes=4 MaxTime=120 AllowGroups=admin
"""

EXAMPLE_SLURMDBD_CONFIG = """#
# `slurmdbd.conf` file generated at 2024-01-30 17:18:36.171652 by slurmutils.
#
ArchiveEvents=yes
ArchiveJobs=yes
ArchiveResvs=yes
ArchiveSteps=no
ArchiveTXN=no
ArchiveUsage=no
ArchiveScript=/usr/sbin/slurm.dbd.archive
AuthInfo=socket=/var/run/munge/munge.socket.2
AuthType=auth/munge
AuthAltTypes=auth/jwt
AuthAltParameters=jwt_key=16549684561684@
DbdHost=slurmdbd-0
DbdBackupHost=slurmdbd-1
DebugLevel=info
PluginDir=/all/these/cool/plugins
PurgeEventAfter=1month
PurgeJobAfter=12month
PurgeResvAfter=1month
PurgeStepAfter=1month
PurgeSuspendAfter=1month
PurgeTXNAfter=12month
PurgeUsageAfter=24month
LogFile=/var/log/slurmdbd.log
PidFile=/var/run/slurmdbd.pid
SlurmUser=slurm
StoragePass=supersecretpasswd
StorageType=accounting_storage/mysql
StorageUser=slurm
StorageHost=127.0.0.1
StoragePort=3306
StorageLoc=slurm_acct_db
"""

EXAMPLE_CGROUP_CONFIG = """#
# `cgroup.conf` file generated at 2024-09-18 15:10:44.652017 by slurmutils.
#
ConstrainCores=yes
ConstrainDevices=yes
ConstrainRAMSpace=yes
ConstrainSwapSpace=yes
"""

EXAMPLE_ACCT_GATHER_CONFIG = """#
# `acct_gather.conf` file generated at 2024-09-18 15:10:44.652017 by slurmutils.
#
EnergyIPMIFrequency=1
EnergyIPMICalcAdjustment=yes
EnergyIPMIPowerSensors=node=16,19;socket1=19,26;knc=16,19
EnergyIPMIUsername=testipmiusername
EnergyIPMIPassword=testipmipassword
EnergyIPMITimeout=10
ProfileHDF5Dir=/mydir
ProfileHDF5Default=all
ProfileInfluxDBDatabase=acct_gather_db
ProfileInfluxDBDefault=all
ProfileInfluxDBHost=testhostname
ProfileInfluxDBPass=testpassword
ProfileInfluxDBRTPolicy=testpolicy
ProfileInfluxDBUser=testuser
ProfileInfluxDBTimeout=10
InfinibandOFEDPort=0
SysfsInterfaces=enp0s1
"""



# Copyright 2024-2025 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities and APIs for interfacing with the Slurm workload manager."""

__all__ = [
    # From `acctgather.py`
    "AcctGatherConfig",
    "AcctGatherConfigEditor",
    # From `cgroup.py`
    "CGroupConfig",
    "CGroupConfigEditor",
    # From `core/*.py`
    "Model",
    "ModelList",
    "ModelMapping",
    "BaseEditor",
    "Callback",
    # From `exceptions.py`
    "ModelError",
    # From `gres.py`
    "Gres",
    "GresConfig",
    "GresConfigEditor",
    "GresList",
    "GresMapping",
    # From `oci.py`
    "OCIConfig",
    "OCIConfigEditor",
    # From `slurm.py`
    "DownNodes",
    "DownNodesList",
    "FrontendNode",
    "FrontendNodeMapping",
    "Node",
    "NodeMapping",
    "NodeSet",
    "NodeSetMapping",
    "Partition",
    "PartitionMapping",
    "SlurmConfig",
    "SlurmConfigEditor",
    # From `slurmdbd.py`
    "SlurmdbdConfig",
    "SlurmdbdConfigEditor",
    # From `utils.py`
    "calculate_rs",
    # Local
    "acctgatherconfig",
    "cgroupconfig",
    "gresconfig",
    "slurmconfig",
    "slurmdbdconfig",
]


from .acctgather import AcctGatherConfig, AcctGatherConfigEditor
from .cgroup import CGroupConfig, CGroupConfigEditor
from .core.base import Model, ModelList, ModelMapping
from .core.callback import Callback
from .core.editor import BaseEditor
from .exceptions import ModelError
from .gres import Gres, GresConfig, GresConfigEditor, GresList, GresMapping
from .oci import OCIConfig, OCIConfigEditor
from .slurm import (
    DownNodes,
    DownNodesList,
    FrontendNode,
    FrontendNodeMapping,
    Node,
    NodeMapping,
    NodeSet,
    NodeSetMapping,
    Partition,
    PartitionMapping,
    SlurmConfig,
    SlurmConfigEditor,
)
from .slurmdbd import SlurmdbdConfig, SlurmdbdConfigEditor
from .utils import calculate_rs

acctgatherconfig = AcctGatherConfigEditor()
cgroupconfig = CGroupConfigEditor()
gresconfig = GresConfigEditor()
ociconfig = OCIConfigEditor()
slurmconfig = SlurmConfigEditor()
slurmdbdconfig = SlurmdbdConfigEditor()
